
import json
from itertools import chain, starmap

import re

from manager.ElasticSearchSourceManager import ElasticSearchSourceManager
from manager.MongoSourceManager import MongoSourceManager

class Facade(object):

    def __init__(self):
        self.srcManager=None
        self.results=[]
        self.numKeys=0


    def flatten_json_iterative_solution(self, dictionary):
        """Flatten a nested json file"""

        def unpack(parent_key, parent_value):
            """Unpack one level of nesting in json file"""
            # Unpack one level only!!!

            if isinstance(parent_value, dict):
                for key, value in parent_value.items():
                    temp1 = parent_key + '_' + key
                    yield temp1, value
            elif isinstance(parent_value, list):
                i = 0
                for value in parent_value:
                    temp2 = parent_key + '_' + str(i)
                    i += 1

                    yield temp2, value
            else:
               yield parent_key, parent_value

                # Keep iterating until the termination condition is satisfied

        while True:
            # Keep unpacking the json file until all values are atomic elements (not dictionary or list)
            dictionary = dict(chain.from_iterable(starmap(unpack, dictionary.items())))
            # Terminate condition: not any value in the json file is dictionary or list
            if not any(isinstance(value, dict) for value in dictionary.values()) and \
                    not any(isinstance(value, list) for value in dictionary.values()):
                break
        return dictionary

    def solve(self,server,port,dbMongo,index,query,keys,source,aggsMongo):
    #def solve(self, server, port, index, query):
        output=list()
        #La prima cosa che facciamo sara quella di creare il manager(connessione)
        if(source=="elastic"):
            numKeys = self.getKeys(query)
            #Serve per poter portare in formato ES la query
            query = self.replaceQuery(query)
            srcManager=ElasticSearchSourceManager(server,port)
            results = srcManager.executeStatement(index, query)
            #se numKeys = 0, il campo _source è vuoto, quindi bisogna controllare che tutti i doc abbiamo il num di chiavi uguali
            if(numKeys == 0):
                listaNumChiavi = []
                for doc in results['hits']['hits']:
                    #vedere se gestire il caso in cui le chiavi sono diverse tra i doc ma il numero uguale per tutti
                    docFlatten=self.flatten_json_iterative_solution(doc['_source'])
                    listaNumChiavi.append(len(list(docFlatten.values())))
                if (len(set(listaNumChiavi)) == 1 and len(listaNumChiavi) > 0):
                    numKeys = listaNumChiavi[0]
                else:
                    return
            for doc in results['hits']['hits']:
                #Nel source ci si aspetta che il programmatore sia preciso quando si tratta di array: es con people,bisogna fare people.nome,people.cognome
                jsonOutputFlatten = self.flatten_json_iterative_solution(doc['_source'])
                #ordiniamo per chiave tutti gli elementi della lista
                jsonOutputFlatten=sorted(jsonOutputFlatten.items(),key=lambda x: x[0])
                jsonOutput=[]
                for i in range(len(jsonOutputFlatten)):
                    jsonOutput.append(jsonOutputFlatten[i][1])
                if (len(jsonOutput) != 0):
                    #le tuple vengono costruite con arieta' numKeys
                    if (len(jsonOutput) % numKeys == 0):
                        i = 0
                        tupla = list()
                        for value in jsonOutput:
                            tupla.append(value)
                            i += 1
                            #se i = numKeys, abbiamo raggiunto l'arita' e bisogna controllare se sono presenti aggregazioni
                            if (i == numKeys):
                                #qui dentro gestiamo le aggregazioni, se sono presenti le appendiamo nella tupla
                                i = 0
                                try:
                                    aggregationsFlatten=self.flatten_json_iterative_solution(results['aggregations'])
                                    aggregationsFlatten = sorted(aggregationsFlatten.items(), key=lambda x: x[0])
                                    #print(aggregationsFlatten)
                                    aggregationsOutput = []
                                    for i in range(len(aggregationsFlatten)):
                                        aggregationsOutput.append(aggregationsFlatten[i][1])
                                    for i in range(len(aggregationsOutput)):
                                        #in caso di valori con la virgola li castiamo ad intero per ovviare al problema del punto in dlv
                                        if (type(aggregationsOutput[i] == float)):
                                            aggregationsOutput[i] = int(aggregationsOutput[i])
                                        tupla.append(aggregationsOutput[i])
                                except KeyError:
                                    pass
                                tupla = tuple(tupla)
                                #escludiamo i duplicati
                                if(tupla not in output):
                                    output.append(tupla)
                                tupla = list()
                    else:
                        #Se il documento non matcha per le chiavi viene eliminato
                        print("Questo documento non matcha per le chiavi richieste------>"+str(jsonOutput))
            output = tuple(output)
            return output
        else:
            query = self.replaceQuery(query)
            numKeys=0
            query = json.loads(query)
            if(keys!=""):
                keys = self.replaceQuery(keys)
                keys=json.loads(keys)
                for temp in keys.values():
                    if (temp == 1):
                        numKeys += 1
            srcManager=MongoSourceManager(server,port,dbMongo)
            results = srcManager.executeStatement(index,query,keys,aggsMongo)
            resultsFlatten=list()
            listaNumChiavi=[]
            for doc in results:
                resultsFlatten.append(doc)
                jsonOutput = self.flatten_json_iterative_solution(doc)
                listaNumChiavi.append(len(jsonOutput.keys()))
            if(len(set(listaNumChiavi))==1):
                if(numKeys==0):
                    numKeys=listaNumChiavi[0]
            else:
                raise "NON POSSO CONTROLLARE QUESTO DOC PERCHE HANNO NUMERO DI CHIAVI DIVERSE"
            #NUMKEYS E' UGUALE A QUESTO
            for doc1 in resultsFlatten:
                jsonOutput = self.flatten_json_iterative_solution(doc1)
                jsonOutput = list(jsonOutput.values())
                if (len(jsonOutput) != 0):
                    if (len(jsonOutput) % numKeys == 0):
                        i = 0
                        tupla = list()
                        for value in jsonOutput:
                            tupla.append(value)
                            i += 1
                            if (i == numKeys):
                                i = 0
                                tupla = tuple(tupla)
                                output.append(tupla)
                                tupla = list()
            output = tuple(output)
            return output
        #Faccio il controllo se tutti i doc hanno le stesse chiavi se non ho messo il source
    def getKeys(self, query):
        numKeys=0
        sourceKeys = re.findall(r"\'_source\'\s*:\s*\[\s*(\'.*\')\s*\]\,*", query)
        if (len(sourceKeys) == 0):
            sourceKeys = re.findall(r"\'_source\':(\'.*?\')", query)
            if (len(sourceKeys) == 0):
                numKeys = 0
            else:
                numKeys = 1
        else:
            numKeys = len(sourceKeys[0].split(","))
        return numKeys
    def replaceQuery(self, query):
        regexKey = re.findall(r"\s*\[*(\'.*?[^\\]\')", query)
        for i in regexKey:
            s = '"' + i[1:-1] + '"'
            query = query.replace(i, s)
        query = query.replace("\\'", "'")
        return query









